package net.codinux.banking.epcqrcode

import com.soywiz.korim.bitmap.Bitmap32
import com.soywiz.korim.format.BMP
import com.soywiz.korim.format.PNG
import com.soywiz.korim.format.TGA
import com.soywiz.korim.qr.QR
import com.soywiz.korim.qr.QRErrorCorrectLevel


open class EpcQrCodeGenerator {

    protected open val qrCode = QR(correctLevel = QRErrorCorrectLevel.M)


    open fun generateEpcQrCode(config: EpcQrCodeConfig): EpcQrCode {
        val heightAndWidth = config.qrCodeHeightAndWidth
        val epcQrCodeContent = generateAsString(config)
        val epcQrCodeContentBytes = epcQrCodeContent.encodeToByteArray() // qrCode.msg() contains a bug for UTF-8 characters -> encode manually to bytes

        val qrCodeBitmap = qrCode.msg(epcQrCodeContentBytes)
            .scaled(heightAndWidth, heightAndWidth, true)

        val qrCodeBytes = convertToFormat(qrCodeBitmap, config.imageFormat)

        return EpcQrCode(qrCodeBitmap, qrCodeBytes, heightAndWidth)
    }

    private fun convertToFormat(qrCodeBitmap: Bitmap32, format: ImageFormat): ByteArray {
        val imageFormat = when (format) {
            ImageFormat.PNG -> PNG
            ImageFormat.TGA -> TGA
            ImageFormat.BMP -> BMP
        }

        return imageFormat.encode(qrCodeBitmap)
    }

    open fun generate(param: EpcQrCodeConfig): EpcQrCodeValues = EpcQrCodeValues(
        EpcQrCodeValues.ServiceTagDefaultValue,
        EpcQrCodeVersion.Version2,
        param.encoding,
        EpcQrCodeValues.FunctionDefaultValue,
        param.bic,
        param.receiverName,
        param.iban,
        EpcQrCodeValues.CurrencyCodeDefaultValue,
        param.amount,
        param.purposeCode,
        null,
        param.reference,
        param.informationForUser
    )

    open fun generateAsString(param: EpcQrCodeConfig): String {
        val epcQrCode = generate(param)

        val epcQrCodeBuilder = StringBuilder()

        epcQrCodeBuilder.appendLine(epcQrCode.serviceTag)
        epcQrCodeBuilder.appendLine(epcQrCode.version.code)
        epcQrCodeBuilder.appendLine(epcQrCode.coding.code)
        epcQrCodeBuilder.appendLine(epcQrCode.function)
        epcQrCodeBuilder.appendLine(mapNullable(epcQrCode.bic))
        epcQrCodeBuilder.appendLine(epcQrCode.receiverName) // TODO: encode to param.encoding
        epcQrCodeBuilder.appendLine(epcQrCode.iban)
        epcQrCodeBuilder.appendLine(mapNullable(epcQrCode.amount?.let { (epcQrCode.currencyCode ?: EpcQrCodeValues.CurrencyCodeDefaultValue) + it })) // TODO: ensure that amount is converted to the correct format
        epcQrCodeBuilder.appendLine(mapNullable(epcQrCode.purposeCode))

        if (epcQrCode.remittanceText != null || epcQrCode.informationForUser != null) {
            epcQrCodeBuilder.appendLine(mapNullable(epcQrCode.remittanceReference))
        } else {
            epcQrCodeBuilder.append(mapNullable(epcQrCode.remittanceReference))
        }

        if (epcQrCode.remittanceText != null || epcQrCode.informationForUser != null) {
            if (epcQrCode.informationForUser != null) {
                epcQrCodeBuilder.appendLine(mapNullable(epcQrCode.remittanceText)) // TODO: encode to param.encoding
            } else {
                epcQrCodeBuilder.append(mapNullable(epcQrCode.remittanceText)) // TODO: encode to param.encoding
            }
        }

        if (epcQrCode.informationForUser != null) {
            epcQrCodeBuilder.append(epcQrCode.informationForUser) // TODO: encode to param.encoding
        }

        return epcQrCodeBuilder.toString()
    }


    protected open fun mapNullable(value: String?): String {
        return value ?: ""
    }

}