package net.codinux.banking.epcqrcode


/**
 * The values of a EPC QR code use two different charsets:
 * - byte = printable ASCII characters from range 32 (white space) to 126 (tilde) (most fields)
 * - characters = all allowed characters according to encoding in line 3 'coding' (only for receiver name, remittance and information for user)
 */
open class EpcQrCodeValues(

    /**
     * 3 bytes.
     * Always(?) has value "BCD".
     */
    open val serviceTag: String,

    /**
     * 3 bytes.
     * Has either value "001" or "002".
     */
    open val version: EpcQrCodeVersion,

    /**
     * 1 byte.
     * The values 1,2,3,4,5,6,7 and 8 determine the interpretation of data to be used.
     * In that order they qualify UTF-8, ISO 8895-1, ISO 8895-2, ISO 8895-4, ISO 8895-5, ISO 8895- 7, ISO 8895-10 and ISO 8895-15
     */
    open val coding: EpcQrCodeCharacterSet,

    /**
     * 3 bytes.
     * The Function is defined by its key value: SCT - SEPA Credit Transfer
     */
    open val function: String,

    /**
     * Receiver's BIC.
     * Mandatory in Version 001, optional in Version 002.
     * Either 8 or 11 bytes.
     */
    open val bic: String?,

    /**
     * Name of the Receiver. Mandatory.
     * Max. 70 characters in encoding of line 3 'coding'.
     */
    open val receiverName: String,

    /**
     * Receiver's IBAN. Mandatory.
     * Max. 34 bytes.
     */
    open val iban: String,

    /**
     * Three capital letter currency code.
     * Only set if amount is also set.
     * The only Currency value available with this version is EUR.
     */
    open val currencyCode: String?,

    /**
     * Optional amount.
     * Max. 12 bytes.
     */
    open val amount: String?,

    /**
     * Optional purpose code.
     * Max. 4 bytes.
     */
    open val purposeCode: String?,

    /**
     * Structured Remittance Information.
     *
     * Creditor Reference (ISO 11649 RF Creditor Reference may be used).
     *
     * May only be set if line 11 (unstructured remittanceText) is empty.
     *
     * Max. 35 bytes
     */
    open val remittanceReference: String?,

    /**
     * Unstructured remittance information (Verwendungstext).
     *
     * May only be set if line 10 (structured remittanceReference) is empty.
     *
     * Max. 140 characters in encoding of line 3 'coding'
     */
    open val remittanceText: String?,

    /**
     * Beneficiary to originator information.
     *
     * The text contained in Display must be shown to the user after decoding and serves a
     * short textual description of what the user is going to initiate.
     * This text must not be forwarded with the data handed over to the payment system.
     *
     * Max. 70 characters in encoding of line 3 'coding'
     */
    open val informationForUser: String?

) {

    companion object {

        const val ServiceTagDefaultValue = "BCD"

        const val FunctionDefaultValue = "SCT"

        const val CurrencyCodeDefaultValue = "EUR"

    }


    /**
     * [remittanceReference] and [remittanceText] are mutual exclusive, that means one of both has to be set
     * but they are never set both at the same time.
     *
     * remittance returns the one that is set.
     */
    open val remittance: String
        get() = remittanceReference ?: remittanceText ?: "" // they should never be both null

    override fun toString(): String {
        return "$receiverName $amount $currencyCode ${remittance}"
    }
}